from __future__ import print_function
import pytest

from xgcm.grid import Grid
from xgcm.test.datasets import datasets_grid_metric


def _expected_result(da, metric, grid, dim, axes, funcname, boundary=None):
    """this is factoring out the expected output of metric aware operations"""
    if funcname == "integrate":
        expected = (da * metric).sum(dim)
    elif funcname == "average":
        expected = (da * metric).sum(dim) / metric.sum(dim)
    elif funcname == "cumint":
        expected = grid.cumsum(da * metric, axes, boundary=boundary)
    else:
        raise ValueError("funcname (`%s`) not recognized" % funcname)
    return expected


@pytest.mark.parametrize("funcname", ["integrate", "average", "cumint"])
@pytest.mark.parametrize(
    "boundary", ["fill", "extend"]
)  # we do not support extrapolate for cumsum?
@pytest.mark.parametrize(
    "periodic",
    [None, "True", "False", {"X": True, "Y": False}, {"X": False, "Y": True}],
)
class TestParametrized:
    def test_bgrid(self, funcname, boundary, periodic):
        ds, coords, metrics = datasets_grid_metric("B")
        grid = Grid(ds, coords=coords, metrics=metrics, periodic=periodic)

        if funcname == "cumint":
            # cumint needs a boundary...
            kwargs = dict(boundary=boundary)
        else:
            # integrate and average do use the boundary input
            kwargs = dict()

        func = getattr(grid, funcname)

        # test tracer position
        for axis, metric_name, dim in zip(
            ["X", "Y", "Z", ["X", "Y"], ["X", "Y", "Z"]],
            ["dx_t", "dy_t", "dz_t", "area_t", "volume_t"],
            ["xt", "yt", "zt", ["xt", "yt"], ["xt", "yt", "zt"]],
        ):
            new = func(ds.tracer, axis, **kwargs)
            expected = _expected_result(
                ds.tracer, ds[metric_name], grid, dim, axis, funcname, **kwargs
            )
            assert new.equals(expected)

            # test with tuple input if list is provided
            if isinstance(axis, list):
                new = func(ds.tracer, tuple(axis), **kwargs)
                assert new.equals(expected)

        # test u position
        for axis, metric_name, dim in zip(
            ["X", "Y", ["X", "Y"]],
            ["dx_ne", "dy_ne", "area_ne"],  # need more metrics?
            ["xu", "yu", ["xu", "yu"]],
        ):
            new = func(ds.u, axis, **kwargs)
            expected = _expected_result(
                ds.u, ds[metric_name], grid, dim, axis, funcname, **kwargs
            )
            assert new.equals(expected)

        # test v position
        for axis, metric_name, dim in zip(
            ["X", "Y", ["X", "Y"]],
            ["dx_ne", "dy_ne", "area_ne"],  # need more metrics?
            ["xu", "yu", ["xu", "yu"]],
        ):
            new = func(ds.v, axis, **kwargs)
            expected = _expected_result(
                ds.v, ds[metric_name], grid, dim, axis, funcname, **kwargs
            )
            assert new.equals(expected)

    def test_cgrid(self, funcname, boundary, periodic):
        ds, coords, metrics = datasets_grid_metric("C")
        grid = Grid(ds, coords=coords, metrics=metrics, periodic=periodic)

        func = getattr(grid, funcname)

        if funcname == "cumint":
            # cumint needs a boundary...
            kwargs = dict(boundary=boundary)
        else:
            # integrate and average do use the boundary input
            kwargs = dict()

        # test tracer position
        for axis, metric_name, dim in zip(
            ["X", "Y", "Z", ["X", "Y"], ["X", "Y", "Z"]],
            ["dx_t", "dy_t", "dz_t", "area_t", "volume_t"],
            ["xt", "yt", "zt", ["xt", "yt"], ["xt", "yt", "zt"]],
        ):
            new = func(ds.tracer, axis, **kwargs)
            expected = _expected_result(
                ds.tracer, ds[metric_name], grid, dim, axis, funcname, **kwargs
            )
            assert new.equals(expected)
            # test with tuple input if list is provided
            if isinstance(axis, list):
                new = func(ds.tracer, tuple(axis), **kwargs)
                assert new.equals(expected)

        # test u positon
        for axis, metric_name, dim in zip(
            ["X", "Y", ["X", "Y"]],
            ["dx_e", "dy_e", "area_e"],  # need more metrics?
            ["xu", "yt", ["xu", "yt"]],
        ):
            new = func(ds.u, axis, **kwargs)
            expected = _expected_result(
                ds.u, ds[metric_name], grid, dim, axis, funcname, **kwargs
            )
            assert new.equals(expected)

        # test v positon
        for axis, metric_name, dim in zip(
            ["X", "Y", ["X", "Y"]],
            ["dx_n", "dy_n", "area_n"],  # need more metrics?
            ["xt", "yu", ["xt", "yu"]],
        ):
            new = func(ds.v, axis, **kwargs)
            expected = _expected_result(
                ds.v, ds[metric_name], grid, dim, axis, funcname, **kwargs
            )
            assert new.equals(expected)

    @pytest.mark.parametrize("axis", ["X", "Y", "Z"])
    def test_missingaxis(self, axis, funcname, periodic, boundary):
        # Error should be raised if application axes
        # include dimension not in datasets

        ds, coords, metrics = datasets_grid_metric("C")

        del coords[axis]

        del_metrics = [k for k in metrics.keys() if axis in k]
        for dm in del_metrics:
            del metrics[dm]

        grid = Grid(ds, coords=coords, metrics=metrics, periodic=periodic)

        func = getattr(grid, funcname)

        if funcname == "cumint":
            # cumint needs a boundary...
            kwargs = dict(boundary=boundary)
        else:
            kwargs = dict()

        match_message = (
            "Unable to find any combinations of metrics for array dims.*%s.*" % axis
        )

        with pytest.raises(KeyError, match=match_message):
            func(ds.tracer, ["X", "Y", "Z"], **kwargs)

        with pytest.raises(KeyError, match=match_message):
            func(ds, axis, **kwargs)

        if axis == "Y":
            # test two missing axes at the same time
            del coords["X"]
            del_metrics = [k for k in metrics.keys() if "X" in k]
            for dm in del_metrics:
                del metrics[dm]

            grid = Grid(ds, coords=coords, metrics=metrics)

            func = getattr(grid, funcname)

            if funcname == "cumint":
                # cumint needs a boundary...
                kwargs = dict(boundary="fill")
            else:
                kwargs = dict()

            match_message = (
                "Unable to find any combinations of metrics for array dims.*X.*Y.*Z.*"
            )
            with pytest.raises(KeyError, match=match_message):
                func(ds, ["X", "Y", "Z"], **kwargs)

            match_message = (
                "Unable to find any combinations of metrics for array dims.*X.*Y.*"
            )
            with pytest.raises(KeyError, match=match_message):
                func(ds, ("X", "Y"), **kwargs)

    def test_missingdim(self, funcname, periodic, boundary):
        ds, coords, metrics = datasets_grid_metric("C")
        grid = Grid(ds, coords=coords, metrics=metrics, periodic=periodic)

        func = getattr(grid, funcname)

        if funcname == "cumint":
            # cumint needs a boundary...
            kwargs = dict(boundary=boundary)
        else:
            kwargs = dict()

        match_message = "Unable to find any combinations of metrics for array dims.*X.*"
        with pytest.raises(KeyError, match=match_message):
            func(ds.tracer.mean("xt"), "X", **kwargs)

        match_message = (
            "Unable to find any combinations of metrics for array dims.*X.*Y.*Z.*"
        )
        with pytest.raises(KeyError, match=match_message):
            func(ds.tracer.mean("xt"), ["X", "Y", "Z"], **kwargs)



from __future__ import print_function
from future.utils import iteritems
import pytest
import xarray as xr
import numpy as np

# example from comodo website
# https://web.archive.org/web/20160417032300/http://pycomodo.forge.imag.fr/norm.html
# netcdf example {
#         dimensions:
#                 ni = 9 ;
#                 ni_u = 10 ;
#         variables:
#                 float ni(ni) ;
#                         ni:axis = "X" ;
#                         ni:standard_name = "x_grid_index" ;
#                         ni:long_name = "x-dimension of the grid" ;
#                         ni:c_grid_dynamic_range = "2:8" ;
#                 float ni_u(ni_u) ;
#                         ni_u:axis = "X" ;
#                         ni_u:standard_name = "x_grid_index_at_u_location" ;
#                         ni_u:long_name = "x-dimension of the grid" ;
#                         ni_u:c_grid_dynamic_range = "3:8" ;
#                         ni_u:c_grid_axis_shift = -0.5 ;
#         data:
#                 ni = 1, 2, 3, 4, 5, 6, 7, 8, 9 ;
#                 ni_u = 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5 ;
# }

N = 100
datasets = {
    # the comodo example, with renamed dimensions
    "1d_outer": xr.Dataset(
        {"data_c": (["XC"], np.random.rand(9)), "data_g": (["XG"], np.random.rand(10))},
        coords={
            "XC": (
                ["XC"],
                np.arange(1, 10),
                {
                    "axis": "X",
                    "standard_name": "x_grid_index",
                    "long_name": "x-dimension of the grid",
                    "c_grid_dynamic_range": "2:8",
                },
            ),
            "XG": (
                ["XG"],
                np.arange(0.5, 10),
                {
                    "axis": "X",
                    "standard_name": "x_grid_index_at_u_location",
                    "long_name": "x-dimension of the grid",
                    "c_grid_dynamic_range": "3:8",
                    "c_grid_axis_shift": -0.5,
                },
            ),
        },
    ),
    "1d_inner": xr.Dataset(
        {"data_c": (["XC"], np.random.rand(9)), "data_g": (["XG"], np.random.rand(8))},
        coords={
            "XC": (
                ["XC"],
                np.arange(1, 10),
                {
                    "axis": "X",
                    "standard_name": "x_grid_index",
                    "long_name": "x-dimension of the grid",
                    "c_grid_dynamic_range": "2:8",
                },
            ),
            "XG": (
                ["XG"],
                np.arange(1.5, 9),
                {
                    "axis": "X",
                    "standard_name": "x_grid_index_at_u_location",
                    "long_name": "x-dimension of the grid",
                    "c_grid_dynamic_range": "3:8",
                    "c_grid_axis_shift": -0.5,
                },
            ),
        },
    ),
    # my own invention
    "1d_left": xr.Dataset(
        {"data_g": (["XG"], np.random.rand(N)), "data_c": (["XC"], np.random.rand(N))},
        coords={
            "XG": (
                ["XG"],
                2 * np.pi / N * np.arange(0, N),
                {"axis": "X", "c_grid_axis_shift": -0.5},
            ),
            "XC": (["XC"], 2 * np.pi / N * (np.arange(0, N) + 0.5), {"axis": "X"}),
        },
    ),
    "1d_right": xr.Dataset(
        {"data_g": (["XG"], np.random.rand(N)), "data_c": (["XC"], np.random.rand(N))},
        coords={
            "XG": (
                ["XG"],
                2 * np.pi / N * np.arange(1, N + 1),
                {"axis": "X", "c_grid_axis_shift": 0.5},
            ),
            "XC": (["XC"], 2 * np.pi / N * (np.arange(0, N) - 0.5), {"axis": "X"}),
        },
    ),
    "2d_left": xr.Dataset(
        {
            "data_g": (["YG", "XG"], np.random.rand(2 * N, N)),
            "data_c": (["YC", "XC"], np.random.rand(2 * N, N)),
        },
        coords={
            "XG": (
                ["XG"],
                2 * np.pi / N * np.arange(0, N),
                {"axis": "X", "c_grid_axis_shift": -0.5},
            ),
            "XC": (["XC"], 2 * np.pi / N * (np.arange(0, N) + 0.5), {"axis": "X"}),
            "YG": (
                ["YG"],
                2 * np.pi / (2 * N) * np.arange(0, 2 * N),
                {"axis": "Y", "c_grid_axis_shift": -0.5},
            ),
            "YC": (
                ["YC"],
                2 * np.pi / (2 * N) * (np.arange(0, 2 * N) + 0.5),
                {"axis": "Y"},
            ),
        },
    ),
}

# include periodicity
datasets_with_periodicity = {
    "nonperiodic_1d_outer": (datasets["1d_outer"], False),
    "nonperiodic_1d_inner": (datasets["1d_inner"], False),
    "periodic_1d_left": (datasets["1d_left"], True),
    "nonperiodic_1d_left": (datasets["1d_left"], False),
    "periodic_1d_right": (datasets["1d_right"], True),
    "nonperiodic_1d_right": (datasets["1d_right"], False),
    "periodic_2d_left": (datasets["2d_left"], True),
    "nonperiodic_2d_left": (datasets["2d_left"], False),
    "xperiodic_2d_left": (datasets["2d_left"], ["X"]),
    "yperiodic_2d_left": (datasets["2d_left"], ["Y"]),
}

expected_values = {
    "nonperiodic_1d_outer": {"axes": {"X": {"center": "XC", "outer": "XG"}}},
    "nonperiodic_1d_inner": {"axes": {"X": {"center": "XC", "inner": "XG"}}},
    "periodic_1d_left": {"axes": {"X": {"center": "XC", "left": "XG"}}},
    "nonperiodic_1d_left": {"axes": {"X": {"center": "XC", "left": "XG"}}},
    "periodic_1d_right": {
        "axes": {"X": {"center": "XC", "right": "XG"}},
        "shift": True,
    },
    "nonperiodic_1d_right": {
        "axes": {"X": {"center": "XC", "right": "XG"}},
        "shift": True,
    },
    "periodic_2d_left": {
        "axes": {
            "X": {"center": "XC", "left": "XG"},
            "Y": {"center": "YC", "left": "YG"},
        }
    },
    "nonperiodic_2d_left": {
        "axes": {
            "X": {"center": "XC", "left": "XG"},
            "Y": {"center": "YC", "left": "YG"},
        }
    },
    "xperiodic_2d_left": {
        "axes": {
            "X": {"center": "XC", "left": "XG"},
            "Y": {"center": "YC", "left": "YG"},
        }
    },
    "yperiodic_2d_left": {
        "axes": {
            "X": {"center": "XC", "left": "XG"},
            "Y": {"center": "YC", "left": "YG"},
        }
    },
}


@pytest.fixture(scope="module", params=datasets_with_periodicity.keys())
def all_datasets(request):
    ds, periodic = datasets_with_periodicity[request.param]
    return ds, periodic, expected_values[request.param]


@pytest.fixture(
    scope="module",
    params=[
        "nonperiodic_1d_outer",
        "nonperiodic_1d_inner",
        "nonperiodic_1d_left",
        "nonperiodic_1d_right",
    ],
)
def nonperiodic_1d(request):
    ds, periodic = datasets_with_periodicity[request.param]
    return ds, periodic, expected_values[request.param]


@pytest.fixture(scope="module", params=["periodic_1d_left", "periodic_1d_right"])
def periodic_1d(request):
    ds, periodic = datasets_with_periodicity[request.param]
    return ds, periodic, expected_values[request.param]


@pytest.fixture(
    scope="module",
    params=[
        "periodic_2d_left",
        "nonperiodic_2d_left",
        "xperiodic_2d_left",
        "yperiodic_2d_left",
    ],
)
def all_2d(request):
    ds, periodic = datasets_with_periodicity[request.param]
    return ds, periodic, expected_values[request.param]


@pytest.fixture(scope="module", params=["periodic_2d_left"])
def periodic_2d(request):
    ds, periodic = datasets_with_periodicity[request.param]
    return ds, periodic, expected_values[request.param]


@pytest.fixture(
    scope="module",
    params=["nonperiodic_2d_left", "xperiodic_2d_left", "yperiodic_2d_left"],
)
def nonperiodic_2d(request):
    ds, periodic = datasets_with_periodicity[request.param]
    return ds, periodic, expected_values[request.param]


def datasets_grid_metric(grid_type):
    """Uniform grid test dataset.
    Should eventually be extended to nonuniform grid"""
    xt = np.arange(4)
    xu = xt + 0.5
    yt = np.arange(5)
    yu = yt + 0.5
    zt = np.arange(6)
    zw = zt + 0.5
    t = np.arange(10)

    def data_generator():
        return np.random.rand(len(xt), len(yt), len(t), len(zt))

    # Need to add a tracer here to get the tracer dimsuffix
    tr = xr.DataArray(
        data_generator(), coords=[("xt", xt), ("yt", yt), ("time", t), ("zt", zt)]
    )

    u_b = xr.DataArray(
        data_generator(), coords=[("xu", xu), ("yu", yu), ("time", t), ("zt", zt)]
    )

    v_b = xr.DataArray(
        data_generator(), coords=[("xu", xu), ("yu", yu), ("time", t), ("zt", zt)]
    )

    u_c = xr.DataArray(
        data_generator(), coords=[("xu", xu), ("yt", yt), ("time", t), ("zt", zt)]
    )

    v_c = xr.DataArray(
        data_generator(), coords=[("xt", xt), ("yu", yu), ("time", t), ("zt", zt)]
    )

    wt = xr.DataArray(
        data_generator(), coords=[("xt", xt), ("yt", yt), ("time", t), ("zw", zw)]
    )

    # maybe also add some other combo of x,t y,t arrays....
    timeseries = xr.DataArray(np.random.rand(len(t)), coords=[("time", t)])

    # northeast distance
    dx = 0.3
    dy = 2
    dz = 20

    dx_ne = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dx - 0.1, coords=[("xu", xu), ("yu", yu)]
    )
    dx_n = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dx - 0.2, coords=[("xt", xt), ("yu", yu)]
    )
    dx_e = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dx - 0.3, coords=[("xu", xu), ("yt", yt)]
    )
    dx_t = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dx - 0.4, coords=[("xt", xt), ("yt", yt)]
    )

    dy_ne = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dy + 0.1, coords=[("xu", xu), ("yu", yu)]
    )
    dy_n = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dy + 0.2, coords=[("xt", xt), ("yu", yu)]
    )
    dy_e = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dy + 0.3, coords=[("xu", xu), ("yt", yt)]
    )
    dy_t = xr.DataArray(
        np.ones([len(xt), len(yt)]) * dy + 0.4, coords=[("xt", xt), ("yt", yt)]
    )

    # dz elements at horizontal tracer points
    dz_t = xr.DataArray(
        data_generator() * dz, coords=[("xt", xt), ("yt", yt), ("time", t), ("zt", zt)]
    )
    dz_w = xr.DataArray(
        data_generator() * dz, coords=[("xt", xt), ("yt", yt), ("time", t), ("zw", zw)]
    )
    # dz elements at velocity points
    dz_w_ne = xr.DataArray(
        data_generator() * dz, coords=[("xu", xu), ("yu", yu), ("time", t), ("zw", zw)]
    )
    dz_w_n = xr.DataArray(
        data_generator() * dz, coords=[("xt", xt), ("yu", yu), ("time", t), ("zw", zw)]
    )
    dz_w_e = xr.DataArray(
        data_generator() * dz, coords=[("xu", xu), ("yt", yt), ("time", t), ("zw", zw)]
    )

    # Make sure the areas are not just the product of x and y distances
    area_ne = (dx_ne * dy_ne) + 0.1
    area_n = (dx_n * dy_n) + 0.2
    area_e = (dx_e * dy_e) + 0.3
    area_t = (dx_t * dy_t) + 0.4

    # calculate volumes, but again add small differences.
    volume_t = (dx_t * dy_t * dz_t) + 0.25

    def _add_metrics(obj):
        obj = obj.copy()
        for name, data in [
            ("dx_ne", dx_ne),
            ("dx_n", dx_n),
            ("dx_e", dx_e),
            ("dx_t", dx_t),
            ("dy_ne", dy_ne),
            ("dy_n", dy_n),
            ("dy_e", dy_e),
            ("dy_t", dy_t),
            ("dz_t", dz_t),
            ("dz_w", dz_w),
            ("dz_w_ne", dz_w_ne),
            ("dz_w_n", dz_w_n),
            ("dz_w_e", dz_w_e),
            ("area_ne", area_ne),
            ("area_n", area_n),
            ("area_e", area_e),
            ("area_t", area_t),
            ("volume_t", volume_t),
        ]:
            obj.coords[name] = data
            obj.coords[name].attrs["tracked_name"] = name
        # add xgcm attrs
        for ii in ["xu", "xt"]:
            obj[ii].attrs["axis"] = "X"
        for ii in ["yu", "yt"]:
            obj[ii].attrs["axis"] = "Y"
        for ii in ["zt", "zw"]:
            obj[ii].attrs["axis"] = "Z"
        for ii in ["time"]:
            obj[ii].attrs["axis"] = "T"
        for ii in ["xu", "yu", "zw"]:
            obj[ii].attrs["c_grid_axis_shift"] = 0.5
        return obj

    coords = {
        "X": {"center": "xt", "right": "xu"},
        "Y": {"center": "yt", "right": "yu"},
        "Z": {"center": "zt", "right": "zw"},
    }

    metrics = {
        ("X",): ["dx_t", "dx_n", "dx_e", "dx_ne"],
        ("Y",): ["dy_t", "dy_n", "dy_e", "dy_ne"],
        ("Z",): ["dz_t", "dz_w", "dz_w_ne", "dz_w_n", "dz_w_e"],
        ("X", "Y"): ["area_t", "area_n", "area_e", "area_ne"],
        ("X", "Y", "Z"): ["volume_t"],
    }

    # combine to different grid configurations (B and C grid)
    if grid_type == "B":
        ds = _add_metrics(
            xr.Dataset(
                {"u": u_b, "v": v_b, "wt": wt, "tracer": tr, "timeseries": timeseries}
            )
        )
    elif grid_type == "C":
        ds = _add_metrics(
            xr.Dataset(
                {"u": u_c, "v": v_c, "wt": wt, "tracer": tr, "timeseries": timeseries}
            )
        )
    else:
        raise ValueError(
            "Invalid input [%s] for `grid_type`. Only supports `B` and `C` at the moment "
            % grid_type
        )

    return ds, coords, metrics



from __future__ import print_function
from future.utils import iteritems
import pytest
import xarray as xr
import numpy as np
from dask.array import from_array

from xgcm.grid import Grid, Axis

from .datasets import (
    all_datasets,
    nonperiodic_1d,
    periodic_1d,
    periodic_2d,
    nonperiodic_2d,
    all_2d,
    datasets,
    datasets_grid_metric,
)


# helper function to produce axes from datasets
def _get_axes(ds):
    all_axes = {ds[c].attrs["axis"] for c in ds.dims if "axis" in ds[c].attrs}
    axis_objs = {ax: Axis(ds, ax) for ax in all_axes}
    return axis_objs


@pytest.mark.parametrize("discontinuity", [None, 10, 360])
@pytest.mark.parametrize("right", [True, False])
def test_extend_right_left(discontinuity, right):
    ds = datasets["1d_left"]
    axis = Axis(ds, "X")
    if discontinuity is None:
        ref = 0
    else:
        ref = discontinuity

    kw = {"boundary_discontinuity": discontinuity}
    if right:
        extended_raw = axis._extend_right(ds.XC, **kw)
        extended = extended_raw[-1]
        expected = ds.XC.data[0] + ref
    else:
        extended_raw = axis._extend_left(ds.XC, **kw)
        extended = extended_raw[0]
        expected = ds.XC.data[-1] - ref
    assert isinstance(extended_raw, np.ndarray)
    assert extended == expected


@pytest.mark.parametrize("fill_value", [0, 10, 20])
@pytest.mark.parametrize("boundary", ["fill", "extend", "extrapolate"])
@pytest.mark.parametrize("periodic", [True, False])
@pytest.mark.parametrize("is_left_edge", [True, False])
@pytest.mark.parametrize("boundary_discontinuity", [None, 360])
def test_get_edge_data(
    periodic, fill_value, boundary, is_left_edge, boundary_discontinuity
):
    ds = datasets["1d_left"]
    axis = Axis(ds, "X", periodic=periodic)
    edge = axis._get_edge_data(
        ds.XC,
        boundary=boundary,
        fill_value=fill_value,
        is_left_edge=is_left_edge,
        boundary_discontinuity=boundary_discontinuity,
    )
    if is_left_edge:
        edge_periodic = ds.XC.data[-1]
        if boundary_discontinuity is not None:
            edge_periodic = edge_periodic - boundary_discontinuity
        edge_extend = ds.XC.data[0]
        edge_extra = ds.XC.data[0] - np.diff(ds.XC.data[0:2])
    else:
        edge_periodic = ds.XC.data[0]
        if boundary_discontinuity is not None:
            edge_periodic = edge_periodic + boundary_discontinuity
        edge_extend = ds.XC.data[-1]
        edge_extra = ds.XC.data[-1] + np.diff(ds.XC.data[-2:])
    edge_fill = fill_value

    if periodic:
        assert edge_periodic == edge
    else:
        if boundary == "fill":
            assert edge_fill == edge
        elif boundary == "extend":
            assert edge_extend == edge
        elif boundary == "extrapolate":
            assert edge_extra == edge
        else:
            assert 0


def test_create_axis(all_datasets):
    ds, periodic, expected = all_datasets
    axis_objs = _get_axes(ds)
    for ax_expected, coords_expected in expected["axes"].items():
        assert ax_expected in axis_objs
        this_axis = axis_objs[ax_expected]
        for axis_name, coord_name in coords_expected.items():
            assert axis_name in this_axis.coords
            assert this_axis.coords[axis_name] == coord_name


def _assert_axes_equal(ax1, ax2):
    assert ax1.name == ax2.name
    for pos, coord in ax1.coords.items():
        assert pos in ax2.coords
        assert coord == ax2.coords[pos]
    assert ax1._periodic == ax2._periodic
    assert ax1._default_shifts == ax2._default_shifts
    assert ax1._facedim == ax2._facedim
    # TODO: make this work...
    # assert ax1._connections == ax2._connections


def test_create_axis_no_comodo(all_datasets):
    ds, periodic, expected = all_datasets
    axis_objs = _get_axes(ds)

    # now strip out the metadata
    ds_noattr = ds.copy()
    for var in ds.variables:
        ds_noattr[var].attrs.clear()

    for axis_name, axis_coords in expected["axes"].items():
        # now create the axis from scratch with no attributes
        ax2 = Axis(ds_noattr, axis_name, coords=axis_coords)
        # and compare to the one created with attributes
        ax1 = axis_objs[axis_name]

        assert ax1.name == ax2.name
        for pos, coord_name in ax1.coords.items():
            assert pos in ax2.coords
            assert coord_name == ax2.coords[pos]
        assert ax1._periodic == ax2._periodic
        assert ax1._default_shifts == ax2._default_shifts
        assert ax1._facedim == ax2._facedim


def test_create_axis_no_coords(all_datasets):
    ds, periodic, expected = all_datasets
    axis_objs = _get_axes(ds)

    ds_drop = ds.drop_vars(list(ds.coords))

    for axis_name, axis_coords in expected["axes"].items():
        # now create the axis from scratch with no attributes OR coords
        ax2 = Axis(ds_drop, axis_name, coords=axis_coords)
        # and compare to the one created with attributes
        ax1 = axis_objs[axis_name]

        assert ax1.name == ax2.name
        for pos, coord in ax1.coords.items():
            assert pos in ax2.coords
        assert ax1._periodic == ax2._periodic
        assert ax1._default_shifts == ax2._default_shifts
        assert ax1._facedim == ax2._facedim


def test_axis_repr(all_datasets):
    ds, periodic, expected = all_datasets
    axis_objs = _get_axes(ds)
    for ax_name, axis in axis_objs.items():
        r = repr(axis).split("\n")
        assert r[0].startswith("<xgcm.Axis")
    # TODO: make this more complete


def test_get_axis_coord(all_datasets):
    ds, periodic, expected = all_datasets
    axis_objs = _get_axes(ds)
    for ax_name, axis in axis_objs.items():
        # create a dataarray with each axis coordinate
        for position, coord in axis.coords.items():
            da = 1 * ds[coord]
            assert axis._get_axis_coord(da) == (position, coord)


def test_axis_wrap_and_replace_2d(periodic_2d):
    ds, periodic, expected = periodic_2d
    axis_objs = _get_axes(ds)

    da_xc_yc = 0 * ds.XC * ds.YC + 1
    da_xc_yg = 0 * ds.XC * ds.YG + 1
    da_xg_yc = 0 * ds.XG * ds.YC + 1

    da_xc_yg_test = axis_objs["Y"]._wrap_and_replace_coords(
        da_xc_yc, da_xc_yc.data, "left"
    )
    assert da_xc_yg.equals(da_xc_yg_test)

    da_xg_yc_test = axis_objs["X"]._wrap_and_replace_coords(
        da_xc_yc, da_xc_yc.data, "left"
    )
    assert da_xg_yc.equals(da_xg_yc_test)


def test_axis_wrap_and_replace_nonperiodic(nonperiodic_1d):
    ds, periodic, expected = nonperiodic_1d
    axis = Axis(ds, "X")

    da_c = 0 * ds.XC + 1
    da_g = 0 * ds.XG + 1

    to = (set(expected["axes"]["X"].keys()) - {"center"}).pop()

    da_g_test = axis._wrap_and_replace_coords(da_c, da_g.data, to)
    assert da_g.equals(da_g_test)

    da_c_test = axis._wrap_and_replace_coords(da_g, da_c.data, "center")
    assert da_c.equals(da_c_test)


# helper functions for padding arrays
# this feels silly...I'm basically just re-coding the function in order to
# test it
def _pad_left(data, boundary, fill_value=0.0):
    pad_val = data[0] if boundary == "extend" else fill_value
    return np.hstack([pad_val, data])


def _pad_right(data, boundary, fill_value=0.0):
    pad_val = data[-1] if boundary == "extend" else fill_value
    return np.hstack([data, pad_val])


@pytest.mark.parametrize(
    "boundary",
    [None, "extend", "fill", pytest.param("extrapolate", marks=pytest.mark.xfail)],
)
@pytest.mark.parametrize("from_center", [True, False])
def test_axis_neighbor_pairs_nonperiodic_1d(nonperiodic_1d, boundary, from_center):
    ds, periodic, expected = nonperiodic_1d
    axis = Axis(ds, "X", periodic=periodic)

    # detect whether this is an outer or inner case
    # outer --> dim_line_diff = 1
    # inner --> dim_line_diff = -1
    dim_len_diff = len(ds.XG) - len(ds.XC)

    if from_center:
        to = (set(expected["axes"]["X"].keys()) - {"center"}).pop()
        da = ds.data_c
    else:
        to = "center"
        da = ds.data_g

    shift = expected.get("shift") or False

    # need boundary condition for everything but outer to center
    if (boundary is None) and (
        dim_len_diff == 0
        or (dim_len_diff == 1 and from_center)
        or (dim_len_diff == -1 and not from_center)
    ):
        with pytest.raises(ValueError):
            data_left, data_right = axis._get_neighbor_data_pairs(
                da, to, boundary=boundary
            )
    else:
        data_left, data_right = axis._get_neighbor_data_pairs(da, to, boundary=boundary)
        if ((dim_len_diff == 1) and not from_center) or (
            (dim_len_diff == -1) and from_center
        ):
            expected_left = da.data[:-1]
            expected_right = da.data[1:]
        elif ((dim_len_diff == 1) and from_center) or (
            (dim_len_diff == -1) and not from_center
        ):
            expected_left = _pad_left(da.data, boundary)
            expected_right = _pad_right(da.data, boundary)
        elif (shift and not from_center) or (not shift and from_center):
            expected_right = da.data
            expected_left = _pad_left(da.data, boundary)[:-1]
        else:
            expected_left = da.data
            expected_right = _pad_right(da.data, boundary)[1:]

        np.testing.assert_allclose(data_left, expected_left)
        np.testing.assert_allclose(data_right, expected_right)


@pytest.mark.parametrize(
    "boundary", ["extend", "fill", pytest.param("extrapolate", marks=pytest.mark.xfail)]
)
def test_axis_cumsum(nonperiodic_1d, boundary):
    ds, periodic, expected = nonperiodic_1d
    axis = Axis(ds, "X", periodic=periodic)

    axis_expected = expected["axes"]["X"]

    cumsum_g = axis.cumsum(ds.data_g, to="center", boundary=boundary)
    assert cumsum_g.dims == ds.data_c.dims
    # check default "to"
    assert cumsum_g.equals(axis.cumsum(ds.data_g, boundary=boundary))

    to = set(axis_expected).difference({"center"}).pop()
    cumsum_c = axis.cumsum(ds.data_c, to=to, boundary=boundary)
    assert cumsum_c.dims == ds.data_g.dims
    # check default "to"
    assert cumsum_c.equals(axis.cumsum(ds.data_c, boundary=boundary))

    cumsum_c_raw = np.cumsum(ds.data_c.data)
    cumsum_g_raw = np.cumsum(ds.data_g.data)

    if to == "right":
        np.testing.assert_allclose(cumsum_c.data, cumsum_c_raw)
        fill_value = 0.0 if boundary == "fill" else cumsum_g_raw[0]
        np.testing.assert_allclose(
            cumsum_g.data, np.hstack([fill_value, cumsum_g_raw[:-1]])
        )
    elif to == "left":
        np.testing.assert_allclose(cumsum_g.data, cumsum_g_raw)
        fill_value = 0.0 if boundary == "fill" else cumsum_c_raw[0]
        np.testing.assert_allclose(
            cumsum_c.data, np.hstack([fill_value, cumsum_c_raw[:-1]])
        )
    elif to == "inner":
        np.testing.assert_allclose(cumsum_c.data, cumsum_c_raw[:-1])
        fill_value = 0.0 if boundary == "fill" else cumsum_g_raw[0]
        np.testing.assert_allclose(cumsum_g.data, np.hstack([fill_value, cumsum_g_raw]))
    elif to == "outer":
        np.testing.assert_allclose(cumsum_g.data, cumsum_g_raw[:-1])
        fill_value = 0.0 if boundary == "fill" else cumsum_c_raw[0]
        np.testing.assert_allclose(cumsum_c.data, np.hstack([fill_value, cumsum_c_raw]))

    ## not much point doing this...we don't have the right test datasets
    ## to really test the errors
    # other_positions = {'left', 'right', 'inner', 'outer'}.difference({to})
    # for pos in other_positions:
    #     with pytest.raises(KeyError):
    #         axis.cumsum(ds.data_c, to=pos, boundary=boundary)


@pytest.mark.parametrize(
    "varname, axis_name, to, roll, roll_axis, swap_order",
    [
        ("data_c", "X", "left", 1, 1, False),
        ("data_c", "Y", "left", 1, 0, False),
        ("data_g", "X", "center", -1, 1, True),
        ("data_g", "Y", "center", -1, 0, True),
    ],
)
def test_axis_neighbor_pairs_2d(
    periodic_2d, varname, axis_name, to, roll, roll_axis, swap_order
):
    ds, periodic, expected = periodic_2d

    axis = Axis(ds, axis_name)

    data = ds[varname]
    data_left, data_right = axis._get_neighbor_data_pairs(data, to)
    if swap_order:
        data_left, data_right = data_right, data_left
    np.testing.assert_allclose(data_left, np.roll(data.data, roll, axis=roll_axis))
    np.testing.assert_allclose(data_right, data.data)


@pytest.mark.parametrize(
    "boundary", ["extend", "fill", pytest.param("extrapolate", marks=pytest.mark.xfail)]
)
@pytest.mark.parametrize("from_center", [True, False])
def test_axis_diff_and_interp_nonperiodic_1d(nonperiodic_1d, boundary, from_center):
    ds, periodic, expected = nonperiodic_1d
    axis = Axis(ds, "X", periodic=periodic)

    dim_len_diff = len(ds.XG) - len(ds.XC)

    if from_center:
        to = (set(expected["axes"]["X"].keys()) - {"center"}).pop()
        coord_to = "XG"
        da = ds.data_c
    else:
        to = "center"
        coord_to = "XC"
        da = ds.data_g

    shift = expected.get("shift") or False

    data = da.data
    if (dim_len_diff == 1 and not from_center) or (dim_len_diff == -1 and from_center):
        data_left = data[:-1]
        data_right = data[1:]
    elif (dim_len_diff == 1 and from_center) or (
        dim_len_diff == -1 and not from_center
    ):
        data_left = _pad_left(data, boundary)
        data_right = _pad_right(data, boundary)
    elif (shift and not from_center) or (not shift and from_center):
        data_left = _pad_left(data[:-1], boundary)
        data_right = data
    else:
        data_left = data
        data_right = _pad_right(data[1:], boundary)

    # interpolate
    data_interp_expected = xr.DataArray(
        0.5 * (data_left + data_right), dims=[coord_to], coords={coord_to: ds[coord_to]}
    )
    data_interp = axis.interp(da, to, boundary=boundary)
    assert data_interp_expected.equals(data_interp)
    # check without "to" specified
    assert data_interp.equals(axis.interp(da, boundary=boundary))

    # difference
    data_diff_expected = xr.DataArray(
        data_right - data_left, dims=[coord_to], coords={coord_to: ds[coord_to]}
    )
    data_diff = axis.diff(da, to, boundary=boundary)
    assert data_diff_expected.equals(data_diff)
    # check without "to" specified
    assert data_diff.equals(axis.diff(da, boundary=boundary))

    # max
    data_max_expected = xr.DataArray(
        np.maximum(data_right, data_left),
        dims=[coord_to],
        coords={coord_to: ds[coord_to]},
    )
    data_max = axis.max(da, to, boundary=boundary)
    assert data_max_expected.equals(data_max)
    # check without "to" specified
    assert data_max.equals(axis.max(da, boundary=boundary))

    # min
    data_min_expected = xr.DataArray(
        np.minimum(data_right, data_left),
        dims=[coord_to],
        coords={coord_to: ds[coord_to]},
    )
    data_min = axis.min(da, to, boundary=boundary)
    assert data_min_expected.equals(data_min)
    # check without "to" specified
    assert data_min.equals(axis.min(da, boundary=boundary))


# this mega test covers all options for 2D data


@pytest.mark.parametrize(
    "boundary", ["extend", "fill", pytest.param("extrapolate", marks=pytest.mark.xfail)]
)
@pytest.mark.parametrize("axis_name", ["X", "Y"])
@pytest.mark.parametrize(
    "varname, this, to", [("data_c", "center", "left"), ("data_g", "left", "center")]
)
def test_axis_diff_and_interp_nonperiodic_2d(
    all_2d, boundary, axis_name, varname, this, to
):
    ds, periodic, expected = all_2d

    try:
        ax_periodic = axis_name in periodic
    except TypeError:
        ax_periodic = periodic

    boundary_arg = boundary if not ax_periodic else None
    axis = Axis(ds, axis_name, periodic=ax_periodic, boundary=boundary_arg)
    da = ds[varname]

    # everything is left shift
    data = ds[varname].data

    axis_num = da.get_axis_num(axis.coords[this])

    # lookups for numpy.pad
    numpy_pad_arg = {"extend": "edge", "fill": "constant"}
    # args for numpy.pad
    pad_left = (1, 0)
    pad_right = (0, 1)
    pad_none = (0, 0)

    if this == "center":
        if ax_periodic:
            data_left = np.roll(data, 1, axis=axis_num)
            data_right = data
        else:
            pad_width = [
                pad_left if i == axis_num else pad_none for i in range(data.ndim)
            ]
            the_slice = tuple(
                [
                    slice(0, -1) if i == axis_num else slice(None)
                    for i in range(data.ndim)
                ]
            )
            data_left = np.pad(data, pad_width, numpy_pad_arg[boundary])[the_slice]
            data_right = data
    elif this == "left":
        if ax_periodic:
            data_left = data
            data_right = np.roll(data, -1, axis=axis_num)
        else:
            pad_width = [
                pad_right if i == axis_num else pad_none for i in range(data.ndim)
            ]
            the_slice = tuple(
                [
                    slice(1, None) if i == axis_num else slice(None)
                    for i in range(data.ndim)
                ]
            )
            data_right = np.pad(data, pad_width, numpy_pad_arg[boundary])[the_slice]
            data_left = data

    data_interp = 0.5 * (data_left + data_right)
    data_diff = data_right - data_left

    # determine new dims
    dims = list(da.dims)
    dims[axis_num] = axis.coords[to]
    coords = {dim: ds[dim] for dim in dims}

    da_interp_expected = xr.DataArray(data_interp, dims=dims, coords=coords)
    da_diff_expected = xr.DataArray(data_diff, dims=dims, coords=coords)

    da_interp = axis.interp(da, to)
    da_diff = axis.diff(da, to)

    assert da_interp_expected.equals(da_interp)
    assert da_diff_expected.equals(da_diff)

    if boundary_arg is not None:
        if boundary == "extend":
            bad_boundary = "fill"
        elif boundary == "fill":
            bad_boundary = "extend"

        da_interp_wrong = axis.interp(da, to, boundary=bad_boundary)
        assert not da_interp_expected.equals(da_interp_wrong)
        da_diff_wrong = axis.diff(da, to, boundary=bad_boundary)
        assert not da_diff_expected.equals(da_diff_wrong)


def test_axis_errors():
    ds = datasets["1d_left"]

    ds_noattr = ds.copy()
    del ds_noattr.XC.attrs["axis"]
    with pytest.raises(
        ValueError, match="Couldn't find a center coordinate for axis X"
    ):
        x_axis = Axis(ds_noattr, "X", periodic=True)

    del ds_noattr.XG.attrs["axis"]
    with pytest.raises(ValueError, match="Couldn't find any coordinates for axis X"):
        x_axis = Axis(ds_noattr, "X", periodic=True)

    ds_chopped = ds.copy().isel(XG=slice(None, 3))
    del ds_chopped["data_g"]
    with pytest.raises(ValueError, match="coordinate XG has incompatible length"):
        x_axis = Axis(ds_chopped, "X", periodic=True)

    ds_chopped.XG.attrs["c_grid_axis_shift"] = -0.5
    with pytest.raises(ValueError, match="coordinate XG has incompatible length"):
        x_axis = Axis(ds_chopped, "X", periodic=True)

    del ds_chopped.XG.attrs["c_grid_axis_shift"]
    with pytest.raises(
        ValueError,
        match="Found two coordinates without `c_grid_axis_shift` attribute for axis X",
    ):
        x_axis = Axis(ds_chopped, "X", periodic=True)

    ax = Axis(ds, "X", periodic=True)

    with pytest.raises(
        ValueError, match="Can't get neighbor pairs for the same position."
    ):
        ax.interp(ds.data_c, "center")

    with pytest.raises(
        ValueError, match="This axis doesn't contain a `right` position"
    ):
        ax.interp(ds.data_c, "right")

    # This case is broken, need to fix!
    # with pytest.raises(
    #    ValueError, match="`boundary=fill` is not allowed " "with periodic axis X."
    # ):
    #    ax.interp(ds.data_c, "left", boundary="fill")


@pytest.mark.parametrize(
    "boundary", [None, "fill", "extend", "extrapolate", {"X": "fill", "Y": "extend"}]
)
@pytest.mark.parametrize("fill_value", [None, 0, 1.0])
def test_grid_create(all_datasets, boundary, fill_value):
    ds, periodic, expected = all_datasets
    grid = Grid(ds, periodic=periodic)
    assert grid is not None
    for ax in grid.axes.values():
        assert ax.boundary is None
    grid = Grid(ds, periodic=periodic, boundary=boundary, fill_value=fill_value)
    for name, ax in grid.axes.items():
        if isinstance(boundary, dict):
            expected = boundary.get(name)
        else:
            expected = boundary
        assert ax.boundary == expected

        if fill_value is None:
            expected = 0.0
        elif isinstance(fill_value, dict):
            expected = fill_value.get(name)
        else:
            expected = fill_value
        assert ax.fill_value == expected


def test_create_grid_no_comodo(all_datasets):
    ds, periodic, expected = all_datasets
    grid_expected = Grid(ds, periodic=periodic)

    ds_noattr = ds.copy()
    for var in ds.variables:
        ds_noattr[var].attrs.clear()

    coords = expected["axes"]
    grid = Grid(ds_noattr, periodic=periodic, coords=coords)

    for axis_name_expected in grid_expected.axes:
        axis_expected = grid_expected.axes[axis_name_expected]
        axis_actual = grid.axes[axis_name_expected]
        _assert_axes_equal(axis_expected, axis_actual)


def test_grid_no_coords(periodic_1d):

    ds, periodic, expected = periodic_1d
    grid_expected = Grid(ds, periodic=periodic)

    ds_nocoords = ds.drop_dims(list(ds.dims.keys()))

    coords = expected["axes"]
    grid = Grid(ds_nocoords, periodic=periodic, coords=coords)

    diff = grid.diff(ds["data_c"], "X")
    assert len(diff.coords) == 0
    interp = grid.interp(ds["data_c"], "X")
    assert len(interp.coords) == 0


def test_grid_repr(all_datasets):
    ds, periodic, expected = all_datasets
    grid = Grid(ds, periodic=periodic)
    r = repr(grid).split("\n")
    assert r[0] == "<xgcm.Grid>"


def test_grid_ops(all_datasets):
    """
    Check that we get the same answer using Axis or Grid objects
    """
    ds, periodic, expected = all_datasets
    grid = Grid(ds, periodic=periodic)

    for axis_name in grid.axes.keys():
        try:
            ax_periodic = axis_name in periodic
        except TypeError:
            ax_periodic = periodic
        axis = Axis(ds, axis_name, periodic=ax_periodic)

        bcs = [None] if ax_periodic else ["fill", "extend"]
        for varname in ["data_c", "data_g"]:
            for boundary in bcs:
                da_interp = grid.interp(ds[varname], axis_name, boundary=boundary)
                da_interp_ax = axis.interp(ds[varname], boundary=boundary)
                assert da_interp.equals(da_interp_ax)
                da_diff = grid.diff(ds[varname], axis_name, boundary=boundary)
                da_diff_ax = axis.diff(ds[varname], boundary=boundary)
                assert da_diff.equals(da_diff_ax)
                if boundary is not None:
                    da_cumsum = grid.cumsum(ds[varname], axis_name, boundary=boundary)
                    da_cumsum_ax = axis.cumsum(ds[varname], boundary=boundary)
                    assert da_cumsum.equals(da_cumsum_ax)


@pytest.mark.parametrize("func", ["interp", "max", "min", "diff", "cumsum"])
@pytest.mark.parametrize("periodic", ["True", "False", ["X"], ["Y"], ["X", "Y"]])
@pytest.mark.parametrize(
    "boundary",
    [
        "fill",
        # "extrapolate", # do we not support extrapolation anymore?
        "extend",
        {"X": "fill", "Y": "extend"},
        {"X": "extend", "Y": "fill"},
    ],
)
def test_multi_axis_input(all_datasets, func, periodic, boundary):
    ds, periodic_unused, expected_unused = all_datasets
    grid = Grid(ds, periodic=periodic)
    axes = list(grid.axes.keys())
    for varname in ["data_c", "data_g"]:
        serial = ds[varname]
        for axis in axes:
            boundary_axis = boundary
            if isinstance(boundary, dict):
                boundary_axis = boundary[axis]
            serial = getattr(grid, func)(serial, axis, boundary=boundary_axis)
        full = getattr(grid, func)(ds[varname], axes, boundary=boundary)
        xr.testing.assert_allclose(serial, full)


def test_grid_dict_input_boundary_fill(nonperiodic_1d):
    """Test axis kwarg input functionality using dict input"""
    ds, _, _ = nonperiodic_1d
    grid_direct = Grid(ds, periodic=False, boundary="fill", fill_value=5)
    grid_dict = Grid(ds, periodic=False, boundary={"X": "fill"}, fill_value={"X": 5})
    assert grid_direct.axes["X"].fill_value == grid_dict.axes["X"].fill_value
    assert grid_direct.axes["X"].boundary == grid_dict.axes["X"].boundary


def test_invalid_boundary_error():
    ds = datasets["1d_left"]
    with pytest.raises(ValueError):
        Axis(ds, "X", boundary="bad")
    with pytest.raises(ValueError):
        Grid(ds, boundary="bad")
    with pytest.raises(ValueError):
        Grid(ds, boundary={"X": "bad"})
    with pytest.raises(ValueError):
        Grid(ds, boundary={"X": 0})
    with pytest.raises(ValueError):
        Grid(ds, boundary=0)


def test_invalid_fill_value_error():
    ds = datasets["1d_left"]
    with pytest.raises(ValueError):
        Axis(ds, "X", fill_value="x")
    with pytest.raises(ValueError):
        Grid(ds, fill_value="bad")
    with pytest.raises(ValueError):
        Grid(ds, fill_value={"X": "bad"})


@pytest.mark.parametrize(
    "funcname",
    [
        "diff",
        "interp",
        "min",
        "max",
        "integrate",
        "average",
        "cumsum",
        "cumint",
        "derivative",
    ],
)
@pytest.mark.parametrize("gridtype", ["B", "C"])
def test_keep_coords(funcname, gridtype):
    ds, coords, metrics = datasets_grid_metric(gridtype)
    ds = ds.assign_coords(yt_bis=ds["yt"], xt_bis=ds["xt"])
    grid = Grid(ds, coords=coords, metrics=metrics)
    func = getattr(grid, funcname)
    for axis_name in grid.axes.keys():
        result = func(ds.tracer, axis_name)
        base_coords = list(result.dims)
        augmented_coords = [
            c for c in ds.tracer.coords if set(ds[c].dims).issubset(result.dims)
        ]
        if funcname in ["integrate", "average"]:
            assert set(result.coords) == set(base_coords + augmented_coords)
        else:
            assert set(result.coords) == set(base_coords)
        #
        if funcname not in ["integrate", "average"]:
            result = func(ds.tracer, axis_name, keep_coords=False)
            assert set(result.coords) == set(base_coords)
            #
            result = func(ds.tracer, axis_name, keep_coords=True)
            assert set(result.coords) == set(base_coords + augmented_coords)


def test_boundary_kwarg_same_as_grid_constructor_kwarg():
    ds = datasets["2d_left"]
    grid1 = Grid(ds, periodic=False)
    grid2 = Grid(ds, periodic=False, boundary={"X": "fill", "Y": "fill"})

    actual1 = grid1.interp(ds.data_g, ("X", "Y"), boundary={"X": "fill", "Y": "fill"})
    actual2 = grid2.interp(ds.data_g, ("X", "Y"))

    xr.testing.assert_identical(actual1, actual2)



"""Compatibility module defining operations on duck numpy-arrays.

Shamelessly copied from xarray."""

from __future__ import division
from __future__ import print_function

import numpy as np

try:
    import dask.array as dsa

    has_dask = True
except ImportError:
    has_dask = False


def _dask_or_eager_func(name, eager_module=np, list_of_args=False, n_array_args=1):
    """Create a function that dispatches to dask for dask array inputs."""
    if has_dask:

        def f(*args, **kwargs):
            dispatch_args = args[0] if list_of_args else args
            if any(isinstance(a, dsa.Array) for a in dispatch_args[:n_array_args]):
                module = dsa
            else:
                module = eager_module
            return getattr(module, name)(*args, **kwargs)

    else:

        def f(data, *args, **kwargs):
            return getattr(eager_module, name)(data, *args, **kwargs)

    return f


insert = _dask_or_eager_func("insert")
take = _dask_or_eager_func("take")
concatenate = _dask_or_eager_func("concatenate", list_of_args=True)


def _apply_boundary_condition(da, dim, left, boundary=None, fill_value=0.0):
    """Supply boundary conditions for  an xarray.DataArray da according along
    the specified dimension. Returns a raw dask or numpy array, depending on
    the underlying data.

    Parameters
    ----------
    da : xarray.DataArray
        The data on which to operate
    dim : str
        Dimenson on which to act
    left : bool
        If `True`, boundary condition is at left (beginning of array).
        If `False`, boundary condition is at the right (end of the array).
    boundary : {'fill', 'extend', 'extrapolate'}
        A flag indicating how the boundary values are determined.

        * 'fill':  All values outside the array set to fill_value
          (i.e. a Neumann boundary condition.)
        * 'extend': Set values outside the array to the nearest array
          value. (i.e. a limited form of Dirichlet boundary condition.)
        * 'extrapolate': Set values by extrapolating linearly from the two
          points nearest to the edge

    fill_value : float, optional
         The value to use in the boundary condition with `boundary='fill'`.
    """

    if boundary not in ["fill", "extend", "extrapolate"]:
        raise ValueError(
            "`boundary` must be 'fill', 'extend' or "
            "'extrapolate', not %r." % boundary
        )

    axis_num = da.get_axis_num(dim)

    # the shape for the edge array
    shape = list(da.shape)
    shape[axis_num] = 1

    base_array = da.data
    index = slice(0, 1) if left else slice(-1, None)
    edge_array = da.isel(**{dim: index}).data

    use_dask = has_dask and isinstance(base_array, dsa.Array)

    if boundary == "extend":
        boundary_array = edge_array
    elif boundary == "fill":
        args = shape, fill_value
        kwargs = {"dtype": base_array.dtype}
        if use_dask:
            full_func = dsa.full
            kwargs["chunks"] = edge_array.chunks
        else:
            full_func = np.full
        boundary_array = full_func(*args, **kwargs)
    elif boundary == "extrapolate":
        gradient_slice = slice(0, 2) if left else slice(-2, None)
        gradient_sign = -1 if left else 1
        linear_gradient = da.isel(**{dim: gradient_slice}).diff(dim=dim).data
        boundary_array = edge_array + gradient_sign * linear_gradient

    return boundary_array


def _pad_array(da, dim, left=False, boundary=None, fill_value=0.0):
    """
    Pad an xarray.DataArray da according to the boundary conditions along dim.
    Return a raw dask or numpy array, depending on the underlying data.

    Parameters
    ----------
    da : xarray.DataArray
        The data on which to operate
    dim : str
        Dimenson to pad
    left : bool
        If `False`, data is padded at the right (end of the array). If `True`,
        padded at left (beginning of array).
    boundary : {'fill', 'extend'}
        A flag indicating how to handle boundaries:

        * None:  Do not apply any boundary conditions. Raise an error if
          boundary conditions are required for the operation.
        * 'fill':  Set values outside the array boundary to fill_value
          (i.e. a Neumann boundary condition.)
        * 'extend': Set values outside the array to the nearest array
          value. (i.e. a limited form of Dirichlet boundary condition.)

    fill_value : float, optional
         The value to use in the boundary condition with `boundary='fill'`.
    """

    if boundary not in ["fill", "extend"]:
        raise ValueError("`boundary` must be `'fill'` or `'extend'`")

    axis_num = da.get_axis_num(dim)
    shape = list(da.shape)
    shape[axis_num] = 1

    base_array = da.data
    index = slice(0, 1) if left else slice(-1, None)
    edge_array = da.isel(**{dim: index}).data

    use_dask = has_dask and isinstance(base_array, dsa.Array)

    if boundary == "extend":
        boundary_array = edge_array
    elif boundary == "fill":
        args = shape, fill_value
        kwargs = {"dtype": base_array.dtype}
        if use_dask:
            full_func = dsa.full
            kwargs["chunks"] = edge_array.chunks
        else:
            full_func = np.full
        boundary_array = full_func(*args, **kwargs)

    arrays_to_concat = [base_array, boundary_array]
    if left:
        arrays_to_concat.reverse()

    return concatenate(arrays_to_concat, axis=axis_num)



import os
import versioneer
from setuptools import find_packages, setup

here = os.path.dirname(__file__)
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

install_requires = [
    "xarray>=0.14.1",
    "dask",
    "numpy",
    "future",
    "docrep<=0.2.7",
]
doc_requires = [
    "sphinx",
    "sphinxcontrib-srclinks",
    "sphinx-pangeo-theme",
    "numpydoc",
    "IPython",
    "nbsphinx",
]

extras_require = {
    "complete": install_requires,
    "docs": doc_requires,
}
extras_require["dev"] = extras_require["complete"] + [
    "pytest",
    "pytest-cov",
    "flake8",
    "black",
    "codecov",
]

setup(
    name="xgcm",
    description="General Circulation Model Postprocessing with xarray",
    url="https://github.com/xgcm/xgcm",
    author="xgcm Developers",
    author_email="rpa@ldeo.columbia.edu",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(exclude=["docs", "tests", "tests.*", "docs.*"]),
    install_requires=install_requires,
    extras_require=extras_require,
    python_requires=">=3.6",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    long_description=long_description,
    long_description_content_type="text/x-rst",
)



from __future__ import print_function
from future.utils import iteritems
import pytest
import xarray as xr
import numpy as np
from xarray.testing import assert_allclose, assert_equal

from xgcm.autogenerate import (
    generate_axis,
    generate_grid_ds,
    _parse_boundary_params,
    _parse_position,
    _position_to_relative,
    _fill_attrs,
)

# create test datasets
pad_value = 1000
dx = 5.0
dy = 2.5
dz = 0.5
a = np.random.rand(int(180 / dy), int(360 / dx), int(10 / dz))
x = np.arange(-180 + dx, 180 + dx, dx)
y = np.arange(-90 + dy, 90 + dy, dy)
z = np.arange(0 + dz, 10 + dz, dz)
x_outer = np.arange(-180 + dx / 2, 180 + dx, dx)
y_outer = np.arange(-90 + dy / 2, 90 + dy, dy)
z_outer = np.arange(0 + dz / 2, 10 + dz, dz)
z_1D = z[0:3]
z_1D_padded = np.hstack([z[0:2] + (dz / 2.0), (z[2] + pad_value) / 2])

xx, yy = np.meshgrid(x, y)
_, _, zz = np.meshgrid(x, y, z)

xx_outer, _ = np.meshgrid(x_outer, y)
_, yy_outer = np.meshgrid(x, y_outer)
_, _, zz_outer = np.meshgrid(x, y, z_outer)

ds_original = xr.Dataset(
    {"somedata": (["lat", "lon", "z"], a)},
    coords={
        "lon": (["lon"], x + (dx / 2.0)),
        "lat": (["lat"], y + (dy / 2.0)),
        "z": (["z"], z + (dz / 2.0)),
        "llon": (["lat", "lon"], xx + (dx / 2.0)),
        "llat": (["lat", "lon"], yy + (dy / 2.0)),
        "zz": (["lat", "lon", "z"], zz + (dz / 2.0)),
    },
)

ds_original_left = xr.Dataset(
    {"somedata": (["lat", "lon", "z"], a)},
    coords={
        "lon": (["lon"], x),
        "lat": (["lat"], y),
        "z": (["z"], z),
        "llon": (["lat", "lon"], xx),
        "llat": (["lat", "lon"], yy),
        "zz": (["lat", "lon", "z"], zz),
    },
)

ds_original_1D = xr.Dataset(
    {"somedata": (["z"], np.array([1, 2, 3]))}, coords={"z": (["z"], z_1D)}
)
ds_original_1D_padded = xr.Dataset(
    {"somedata": (["z"], np.array([1, 2, 3]))},
    coords={"z": (["z"], z[0:3]), "test": (["test"], z_1D_padded)},
)
ds_out_left = xr.Dataset(
    {"somedata": (["lat", "lon", "z"], a)},
    coords={
        "lon": (["lon"], x + (dx / 2.0), {"axis": "X"}),
        "lat": (["lat"], y + (dy / 2.0), {"axis": "Y"}),
        "z": (["z"], z + (dz / 2.0), {"axis": "Z"}),
        "llon": (["lat", "lon"], xx + (dx / 2.0), {"axis": "X"}),
        "llat": (["lat", "lon"], yy + (dy / 2.0), {"axis": "Y"}),
        "zz": (["lat", "lon", "z"], zz + (dz / 2.0), {"axis": "Z"}),
        "lon_left": (["lon_left"], x, {"axis": "X", "c_grid_axis_shift": -0.5}),
        "lat_left": (["lat_left"], y, {"axis": "Y", "c_grid_axis_shift": -0.5}),
        "z_left": (["z_left"], z, {"axis": "Z", "c_grid_axis_shift": -0.5}),
        "llon_left": (
            ["lat", "lon_left"],
            xx,
            {"axis": "X", "c_grid_axis_shift": -0.5},
        ),
        "llat_left": (
            ["lat_left", "lon"],
            yy,
            {"axis": "Y", "c_grid_axis_shift": -0.5},
        ),
        "zz_left": (
            ["lat", "lon", "z_left"],
            zz,
            {"axis": "Z", "c_grid_axis_shift": -0.5},
        ),
    },
)

ds_out_right = xr.Dataset(
    {"somedata": (["lat", "lon", "z"], a)},
    coords={
        "lon": (["lon"], x + (dx / 2.0), {"axis": "X", "c_grid_axis_shift": -0.5}),
        "lat": (["lat"], y + (dy / 2.0), {"axis": "Y", "c_grid_axis_shift": -0.5}),
        "z": (["z"], z + (dz / 2.0), {"axis": "Z", "c_grid_axis_shift": -0.5}),
        "llon": (
            ["lat", "lon"],
            xx + (dx / 2.0),
            {"axis": "X", "c_grid_axis_shift": -0.5},
        ),
        "llat": (
            ["lat", "lon"],
            yy + (dy / 2.0),
            {"axis": "Y", "c_grid_axis_shift": -0.5},
        ),
        "zz": (
            ["lat", "lon", "z"],
            zz + (dz / 2.0),
            {"axis": "Z", "c_grid_axis_shift": -0.5},
        ),
        "lon_right": (["lon_right"], x + dx, {"axis": "X"}),
        "lat_right": (["lat_right"], y + dy, {"axis": "Y"}),
        "z_right": (["z_right"], z + dz, {"axis": "Z"}),
        "llon_right": (["lat", "lon_right"], xx + dx, {"axis": "X"}),
        "llat_right": (["lat_right", "lon"], yy + dy, {"axis": "Y"}),
        "zz_right": (["lat", "lon", "z_right"], zz + dz, {"axis": "Z"}),
    },
)

ds_out_center = xr.Dataset(
    {"somedata": (["lat", "lon", "z"], a)},
    coords={
        "lon": (["lon"], x, {"axis": "X", "c_grid_axis_shift": -0.5}),
        "lat": (["lat"], y, {"axis": "Y", "c_grid_axis_shift": -0.5}),
        "z": (["z"], z, {"axis": "Z", "c_grid_axis_shift": -0.5}),
        "llon": (["lat", "lon"], xx, {"axis": "X", "c_grid_axis_shift": -0.5}),
        "llat": (["lat", "lon"], yy, {"axis": "Y", "c_grid_axis_shift": -0.5}),
        "zz": (["lat", "lon", "z"], zz, {"axis": "Z", "c_grid_axis_shift": -0.5}),
        "lon_center": (["lon_center"], x + (dx / 2.0), {"axis": "X"}),
        "lat_center": (["lat_center"], y + (dy / 2.0), {"axis": "Y"}),
        "z_center": (["z_center"], z + (dz / 2.0), {"axis": "Z"}),
        "llon_center": (["lat", "lon_center"], xx + (dx / 2.0), {"axis": "X"}),
        "llat_center": (["lat_center", "lon"], yy + (dy / 2.0), {"axis": "Y"}),
        "zz_center": (["lat", "lon", "z_center"], zz + (dz / 2.0), {"axis": "Z"}),
    },
)

ds_out_outer = xr.Dataset(
    {"somedata": (["lat", "lon", "z"], a)},
    coords={
        "lon": (["lon"], x, {"axis": "X"}),
        "lat": (["lat"], y, {"axis": "Y"}),
        "z": (["z"], z, {"axis": "Z"}),
        "llon": (["lat", "lon"], xx, {"axis": "X"}),
        "llat": (["lat", "lon"], yy, {"axis": "Y"}),
        "zz": (["lat", "lon", "z"], zz, {"axis": "Z"}),
        "lon_outer": (["lon_outer"], x_outer, {"axis": "X", "c_grid_axis_shift": -0.5}),
        "lat_outer": (["lat_outer"], y_outer, {"axis": "Y", "c_grid_axis_shift": -0.5}),
        "z_outer": (["z_outer"], z_outer, {"axis": "Z", "c_grid_axis_shift": -0.5}),
        "llon_outer": (
            ["lat", "lon_outer"],
            xx_outer,
            {"axis": "X", "c_grid_axis_shift": -0.5},
        ),
        "llat_outer": (
            ["lat_outer", "lon"],
            yy_outer,
            {"axis": "Y", "c_grid_axis_shift": -0.5},
        ),
        "zz_outer": (
            ["lat", "lon", "z_outer"],
            zz_outer,
            {"axis": "Z", "c_grid_axis_shift": -0.5},
        ),
    },
)


def test_generate_axis():
    a = generate_axis(
        ds_original,
        "X",
        "lon",
        "lon",
        pos_from="center",
        pos_to="right",
        pad=None,
        boundary_discontinuity=360,
    )
    b = generate_axis(
        ds_original,
        "Y",
        "lat",
        "lat",
        pos_from="center",
        pos_to="left",
        pad=None,
        boundary_discontinuity=180,
    )
    c = generate_axis(
        ds_original, "Z", "z", "z", pos_from="center", pos_to="left", pad="auto"
    )
    d = generate_axis(
        ds_original_1D,
        "Z",
        "z",
        "z",
        pos_from="left",
        pos_to="center",
        pad=pad_value,
        new_name="test",
    )
    e = generate_axis(
        ds_original_left, "Z", "z", "z", pos_from="left", pos_to="center", pad="auto"
    )
    f = generate_axis(
        ds_original_left, "Z", "z", "z", pos_from="center", pos_to="outer", pad="auto"
    )
    g = generate_axis(
        ds_original_left,
        "X",
        "lon",
        "lon",
        pos_from="center",
        pos_to="outer",
        pad=None,
        boundary_discontinuity=360,
    )

    assert_allclose(a["lon_right"], ds_out_right["lon_right"])
    assert_allclose(b["lat_left"], ds_out_left["lat_left"])
    assert_allclose(c["z_left"], ds_out_left["z_left"])
    assert_allclose(d["test"], ds_original_1D_padded["test"])
    assert_allclose(e["z_center"], ds_out_center["z_center"])
    assert_allclose(f["z_outer"], ds_out_outer["z_outer"])
    assert_allclose(g["lon_outer"], ds_out_outer["lon_outer"])

    # Mulitdim cases
    aa = generate_axis(
        a,
        "X",
        "llon",
        "lon",
        pos_from="center",
        pos_to="right",
        pad=None,
        boundary_discontinuity=360,
        attrs_from_scratch=False,
    )
    bb = generate_axis(
        b,
        "Y",
        "llat",
        "lat",
        pos_from="center",
        pos_to="left",
        pad=None,
        boundary_discontinuity=180,
        attrs_from_scratch=False,
    )
    ee = generate_axis(
        e,
        "Z",
        "zz",
        "z",
        pos_from="left",
        pos_to="center",
        pad="auto",
        attrs_from_scratch=False,
    )
    ff = generate_axis(
        f,
        "Z",
        "zz",
        "z",
        pos_from="center",
        pos_to="outer",
        pad="auto",
        attrs_from_scratch=False,
    )

    gg = generate_axis(
        g,
        "X",
        "llon",
        "lon",
        pos_from="center",
        pos_to="outer",
        pad=None,
        boundary_discontinuity=360,
        attrs_from_scratch=False,
    )

    assert_allclose(aa["llon_right"], ds_out_right["llon_right"])
    assert_allclose(bb["llat_left"], ds_out_left["llat_left"])
    assert_allclose(ee["zz_center"], ds_out_center["zz_center"])
    assert_allclose(ff["zz_outer"], ds_out_outer["zz_outer"])
    assert_allclose(gg["llon_outer"], ds_out_outer["llon_outer"])

    with pytest.raises(ValueError):
        # Check if generate axis fails when a DataArray is passed instead of
        # Dataset
        generate_axis(
            c["somedata"],
            "Z",
            "zz",
            "z",
            pos_from="left",
            pos_to="center",
            pad="auto",
            attrs_from_scratch=False,
        )
    with pytest.raises(ValueError):
        generate_axis(c, "Z", "zz", "z", pad="auto", boundary_discontinuity=360)
    with pytest.raises(ValueError):
        generate_axis(c, "Z", "zz", "z", pad=None, boundary_discontinuity=None)


def test_generate_grid_ds():
    # This needs more cases
    axis_dims = {"X": "lon", "Y": "lat", "Z": "z"}
    axis_coords = {"X": "llon", "Y": "llat", "Z": "zz"}
    position = ("center", "outer")
    boundary_discontinuity = {"lon": 360, "llon": 360, "lat": 180, "llat": 180}
    pad = {"z": "auto", "zz": "auto"}
    ds = generate_grid_ds(
        ds_original_left, axis_dims, axis_coords, position, boundary_discontinuity, pad
    )
    assert_equal(ds, ds_out_outer)


def test_parse_boundary_params():
    assert _parse_boundary_params(360, "anything") == 360
    assert _parse_boundary_params({"something": 360}, "something") == 360
    assert _parse_boundary_params({"something": 360}, "something_else") is None


@pytest.mark.parametrize(
    "p_f, p_t",
    [("left", "center"), ("center", "left"), ("center", "right"), ("right", "center")],
)
def test_parse_position(p_f, p_t):
    default = ("center", "left")
    assert _parse_position((p_f, p_t), "anything") == (p_f, p_t)
    assert _parse_position({"a": (p_f, p_t)}, "a") == (p_f, p_t)
    assert _parse_position({"a": (p_f, p_t)}, "b") == default
    assert _parse_position({"a": (p_f, p_t), "b": (p_t, p_f)}, "a") == (p_f, p_t)
    assert _parse_position({"a": (p_f, p_t), "b": (p_t, p_f)}, "b") == (p_t, p_f)


@pytest.mark.parametrize(
    "p, relative",
    [
        (("left", "center"), "right"),
        (("center", "left"), "left"),
        (("center", "right"), "right"),
        (("right", "center"), "left"),
        (("center", "outer"), "outer"),
        (("center", "inner"), "inner"),
    ],
)
def test_position_to_relative(p, relative):
    assert _position_to_relative(p[0], p[1]) == relative
    error_test = [
        ("left", "right"),
        ("inner", "outer"),
        ("left", "outer"),
        ("right", "outer"),
        ("left", "inner"),
        ("right", "inner"),
    ]
    for et in error_test:
        with pytest.raises(RuntimeError):
            _position_to_relative(et[0], et[1])
        with pytest.raises(RuntimeError):
            _position_to_relative(et[1], et[0])


def test_fill_attrs():
    a = _fill_attrs(ds_out_right["lon"], "right", "X")
    assert a.attrs["axis"] == "X"
    assert a.attrs["c_grid_axis_shift"] == 0.5

    a = _fill_attrs(ds_out_right["lon"], "left", "Z")
    assert a.attrs["axis"] == "Z"
    assert a.attrs["c_grid_axis_shift"] == -0.5

    a = _fill_attrs(ds_out_right["lon"], "center", "Y")
    assert a.attrs["axis"] == "Y"
    assert "c_grid_axis_shift" not in a.attrs.keys()



from __future__ import print_function

import numpy as np
import dask as dsk
from xgcm.duck_array_ops import concatenate


def test_concatenate():
    a = np.array([1, 2, 3])
    b = np.array([10])
    a_dask = dsk.array.from_array(a, chunks=1)
    b_dask = dsk.array.from_array(b, chunks=1)
    concat = concatenate([a, b], axis=0)
    concat_dask = concatenate([a_dask, b_dask], axis=0)
    concat_mixed = concatenate([a, b_dask], axis=0)
    assert isinstance(concat, np.ndarray)
    assert isinstance(concat_dask, dsk.array.Array)
    assert isinstance(concat_mixed, dsk.array.Array)



from __future__ import print_function
from future.utils import iteritems
import pytest
import xarray as xr
import numpy as np
from dask.array import from_array

from xgcm.grid import Grid, Axis


@pytest.fixture(scope="module")
def ds():
    N = 25
    return xr.Dataset(
        {
            "data_c": (["face", "y", "x"], np.random.rand(2, N, N)),
            "u": (["face", "xl", "y"], np.random.rand(2, N, N)),
            "v": (["face", "x", "yl"], np.random.rand(2, N, N)),
        },
        coords={
            "x": (("x",), np.arange(N), {"axis": "X"}),
            "xl": (
                ("xl"),
                np.arange(N) - 0.5,
                {"axis": "X", "c_grid_axis_shift": -0.5},
            ),
            "y": (("y",), np.arange(N), {"axis": "Y"}),
            "yl": (
                ("yl"),
                np.arange(N) - 0.5,
                {"axis": "Y", "c_grid_axis_shift": -0.5},
            ),
            "face": (("face",), [0, 1]),
        },
    )


# ---- structure of face_connections dictionaries ----
# key: index of face
# value: another dictionary
#   key: axis name
#   value: a tuple of link specifiers
#        neighbor face index,
#          neighboring axis to connect to,
#            whether to reverse the connection


@pytest.fixture(scope="module")
def ds_face_connections_x_to_x():
    return {
        "face": {0: {"X": (None, (1, "X", False))}, 1: {"X": ((0, "X", False), None)}}
    }


@pytest.fixture(scope="module")
def ds_face_connections_x_to_y():
    return {
        "face": {0: {"X": (None, (1, "Y", False))}, 1: {"Y": ((0, "X", False), None)}}
    }


@pytest.fixture(scope="module")
def cs():
    # cubed-sphere
    N = 25
    ds = xr.Dataset(
        {"data_c": (["face", "y", "x"], np.random.rand(6, N, N))},
        coords={
            "x": (("x",), np.arange(N), {"axis": "X"}),
            "xl": (
                ("xl"),
                np.arange(N) - 0.5,
                {"axis": "X", "c_grid_axis_shift": -0.5},
            ),
            "y": (("y",), np.arange(N), {"axis": "Y"}),
            "yl": (
                ("yl"),
                np.arange(N) - 0.5,
                {"axis": "Y", "c_grid_axis_shift": -0.5},
            ),
            "face": (("face",), np.arange(6)),
        },
    )
    return ds


# TODO: consider revising this to avoid any reversed connections, which
# can cause problems for vector interpolation
@pytest.fixture(scope="module")
def cubed_sphere_connections():
    return {
        "face": {
            0: {
                "X": ((3, "X", False), (1, "X", False)),
                "Y": ((4, "Y", False), (5, "Y", False)),
            },
            1: {
                "X": ((0, "X", False), (2, "X", False)),
                "Y": ((4, "X", False), (5, "X", True)),
            },
            2: {
                "X": ((1, "X", False), (3, "X", False)),
                "Y": ((4, "Y", True), (5, "Y", True)),
            },
            3: {
                "X": ((2, "X", False), (0, "X", False)),
                "Y": ((4, "X", True), (5, "X", False)),
            },
            4: {
                "X": ((3, "Y", True), (1, "Y", False)),
                "Y": ((2, "Y", True), (0, "Y", False)),
            },
            5: {
                "X": ((3, "Y", False), (1, "Y", True)),
                "Y": ((0, "Y", False), (2, "Y", True)),
            },
        }
    }


def test_create_periodic_grid(ds):
    ds = ds.isel(face=0)
    grid = Grid(ds, periodic=True)
    for axis_name in grid.axes:
        axis = grid.axes[axis_name]
        assert axis._facedim is None
        connect_left, connect_right = axis._connections[None]
        assert connect_left[0] is None
        assert connect_right[0] is None
        assert connect_left[1] is axis
        assert connect_right[1] is axis
        assert connect_left[2] is False
        assert connect_right[2] is False


def test_get_periodic_grid_edge(ds):
    ds = ds.isel(face=0)
    grid = Grid(ds, periodic=True)

    xaxis = grid.axes["X"]
    left_edge_data_x = xaxis._get_edge_data(ds.data_c)
    np.testing.assert_allclose(left_edge_data_x, ds.data_c.isel(x=-1).data[:, None])
    right_edge_data_x = xaxis._get_edge_data(ds.data_c, is_left_edge=False)
    np.testing.assert_allclose(right_edge_data_x, ds.data_c.isel(x=0).data[:, None])

    yaxis = grid.axes["Y"]
    left_edge_data_y = yaxis._get_edge_data(ds.data_c)
    np.testing.assert_allclose(left_edge_data_y, ds.data_c.isel(y=-1).data[None, :])
    right_edge_data_y = yaxis._get_edge_data(ds.data_c, is_left_edge=False)
    np.testing.assert_allclose(right_edge_data_y, ds.data_c.isel(y=0).data[None, :])


def test_connection_errors(ds):
    pass


def test_create_connected_grid(ds, ds_face_connections_x_to_x):
    # simplest scenario with one face connection
    grid = Grid(ds, face_connections=ds_face_connections_x_to_x)

    xaxis = grid.axes["X"]
    yaxis = grid.axes["Y"]

    # make sure we have actual axis objects in the connection dict
    # this is a bad test because it tests the details of the implementation,
    # not the behavior. But it is useful for now
    assert xaxis._facedim == "face"
    assert xaxis._connections[0][1][0] == 1
    assert xaxis._connections[0][1][1] is xaxis
    assert xaxis._connections[1][0][0] == 0
    assert xaxis._connections[1][0][1] is xaxis


def test_diff_interp_connected_grid_x_to_x(ds, ds_face_connections_x_to_x):
    # simplest scenario with one face connection
    grid = Grid(ds, face_connections=ds_face_connections_x_to_x)
    diff_x = grid.diff(ds.data_c, "X", boundary="fill")
    interp_x = grid.interp(ds.data_c, "X", boundary="fill")

    # make sure the face connection got applied correctly
    np.testing.assert_allclose(
        diff_x[1, :, 0], ds.data_c[1, :, 0] - ds.data_c[0, :, -1]
    )
    np.testing.assert_allclose(
        interp_x[1, :, 0], 0.5 * (ds.data_c[1, :, 0] + ds.data_c[0, :, -1])
    )

    # make sure the left boundary got applied correctly
    np.testing.assert_allclose(diff_x[0, :, 0], ds.data_c[0, :, 0] - 0.0)
    np.testing.assert_allclose(interp_x[0, :, 0], 0.5 * (ds.data_c[0, :, 0] + 0.0))


def test_diff_interp_connected_grid_x_to_y(ds, ds_face_connections_x_to_y):
    # one face connection, rotated
    grid = Grid(ds, face_connections=ds_face_connections_x_to_y)

    diff_x = grid.diff(ds.data_c, "X", boundary="fill")
    interp_x = grid.interp(ds.data_c, "X", boundary="fill")
    diff_y = grid.diff(ds.data_c, "Y", boundary="fill")
    interp_y = grid.interp(ds.data_c, "Y", boundary="fill")

    # make sure the face connection got applied correctly
    # non-same axis connections require rotation
    # ravel everything to avoid dealing with broadcasting
    np.testing.assert_allclose(
        diff_y.data[1, 0, :].ravel(),
        ds.data_c.data[1, 0, :].ravel() - ds.data_c.data[0, ::-1, -1].ravel(),
    )

    np.testing.assert_allclose(
        interp_y.data[1, 0, :].ravel(),
        0.5 * (ds.data_c.data[1, 0, :].ravel() + ds.data_c.data[0, ::-1, -1].ravel()),
    )

    # TODO: checking all the other boundaries


def test_vector_diff_interp_connected_grid_x_to_y(ds, ds_face_connections_x_to_y):
    # simplest scenario with one face connection
    grid = Grid(ds, face_connections=ds_face_connections_x_to_y)

    # interp u and v to cell center
    vector_center = grid.interp_2d_vector(
        {"X": ds.u, "Y": ds.v}, to="center", boundary="fill"
    )
    u_c_interp, v_c_interp = vector_center["X"], vector_center["Y"]

    vector_diff = grid.diff_2d_vector(
        {"X": ds.u, "Y": ds.v}, to="center", boundary="fill"
    )
    u_c_diff, v_c_diff = vector_diff["X"], vector_diff["Y"]

    # first point should be normal
    np.testing.assert_allclose(
        u_c_interp.data[0, 0, :], 0.5 * (ds.u.data[0, 0, :] + ds.u.data[0, 1, :])
    )
    np.testing.assert_allclose(
        u_c_diff.data[0, 0, :], ds.u.data[0, 1, :] - ds.u.data[0, 0, :]
    )

    # last point should be fancy
    np.testing.assert_allclose(
        u_c_interp.data[0, -1, :], 0.5 * (ds.u.data[0, -1, :] + ds.v.data[1, ::-1, 0])
    )
    np.testing.assert_allclose(
        u_c_diff.data[0, -1, :], -ds.u.data[0, -1, :] + ds.v.data[1, ::-1, 0]
    )

    # TODO: figure out tangent vectors
    with pytest.raises(NotImplementedError):
        vector_corner = grid.interp_2d_vector(
            {"X": ds.v, "Y": ds.u}, to="left", boundary="fill"
        )
    with pytest.raises(NotImplementedError):
        vector_corner = grid.interp_2d_vector({"X": ds.v, "Y": ds.u}, boundary="fill")


def test_create_cubed_sphere_grid(cs, cubed_sphere_connections):
    grid = Grid(cs, face_connections=cubed_sphere_connections)


def test_diff_interp_cubed_sphere(cs, cubed_sphere_connections):
    grid = Grid(cs, face_connections=cubed_sphere_connections)
    face, _ = xr.broadcast(cs.face, cs.data_c)

    face_diff_x = grid.diff(face, "X")
    np.testing.assert_allclose(face_diff_x[:, 0, 0], [-3, 1, 1, 1, 1, 2])
    np.testing.assert_allclose(face_diff_x[:, -1, 0], [-3, 1, 1, 1, 1, 2])

    face_diff_y = grid.diff(face, "Y")
    np.testing.assert_allclose(face_diff_y[:, 0, 0], [-4, -3, -2, -1, 2, 5])
    np.testing.assert_allclose(face_diff_y[:, 0, -1], [-4, -3, -2, -1, 2, 5])
